<?php
#===============================================================================
#	$Id: thumbcache.php,v 1.18 2013/04/21 17:59:18 wolf Exp $
#===============================================================================
#
#	Provides Thumbnail-JPEG for Image-URL
#
#	image=		Image or Website URL. Mandatory.
#	DEBUG=		If set, error messages instead of replacement picture
#
#===============================================================================

$THUMB_WIDTH	= 120;
$THUMB_HEIGHT	= 120;
$THUMB_MINRATIO	= 0.4;
$THUMB_MAXRATIO	= 2.0;

$IMAGE_POS_CACHE_TIME	= 7 * 24 * 3600;	# 1 week
$IMAGE_NEG_CACHE_TIME	= 1 *  6 * 3600;	# 6 hours
$ROBOTS_CACHE_TIME	= 1 * 24 * 3600;	# 1 day

$ALLOWED_DOMAINS = array (
	"flickr.com",
	"openstreetmap.org",
	"staticflickr.com",
	"wikimedia.org",
	"wikipedia.org",
);

$FLICKR_THUMB_SIZE = "m";

$ERROR_IMAGE	= '/i/w/sorry_160x80.gif';
$ERROR_MIME	= 'image/jpeg';
$CACHE_DIR	= '/tmp';

$USER_AGENT = 'thumbcache.php @ http://www.historic.place/; prereads, thumbnails and caches images to display in map popup info boxes. In case of trouble contact <thumbnail@example.org>.';

#-------------------------------------------------------------------------------
#	Global values
#-------------------------------------------------------------------------------

$DOCUMENT_ROOT	= preg_replace ('#/+$#', '', $_SERVER['DOCUMENT_ROOT']);
$NO_CACHE	= @$_SERVER['HTTP_PRAGMA']=="no-cache" ||
			@$_SERVER['HTTP_CACHE_CONTROL']=='no-cache';
$ROBOT_FRESH	= time() - ($NO_CACHE ? 0 : $ROBOTS_CACHE_TIME);
$DEBUG		= @intval(@$_GET['DEBUG']) + 0;

#-------------------------------------------------------------------------------
#	Wikimedia commons requires a non default User-Agent
#-------------------------------------------------------------------------------

ini_set('user_agent', $USER_AGENT);

#-------------------------------------------------------------------------------
#	Logger
#-------------------------------------------------------------------------------

$lapTime = microtime(TRUE);

function usedTime () {

	global $lapTime;

	$startTime = $lapTime;
	$lapTime = microtime(TRUE);
	return sprintf ('%.3f', $lapTime - $startTime);
}

$LOG = array();

function logMsg ($msg) {

	global $LOG, $DEBUG;

	$LOG[] = $msg;

	if ($DEBUG >= 3) error_log($msg);
}

#-------------------------------------------------------------------------------
#	Error handler
#-------------------------------------------------------------------------------

function Fail ($message) {

	global $LOG;
	global $DEBUG;
	global $DOCUMENT_ROOT;

	header ("HTTP/1.0 400 Something went wrong.");

	#-----------------------------------------------------------------------
	#	Replacement graphic
	#-----------------------------------------------------------------------

	if (!$DEBUG) {

		global $ERROR_IMAGE;
		global $ERROR_MIME;

		$path = "{$DOCUMENT_ROOT}/{$ERROR_IMAGE}";
		$size = @filesize($path);

		if ($size) {

			header ("Content-Type: {$ERROR_MIME}");
			header ("Content-Length: {$size}");

			readfile ($path);

			exit (1);
		}
	}

	#-----------------------------------------------------------------------
	#	Debug mode
	#-----------------------------------------------------------------------

	header ("Content-Type: text/plain");

	echo implode ("\n", $LOG),"\n";
	if (count($LOG)) echo "\n";
	echo "$message\n";

	exit (0);
}

#-------------------------------------------------------------------------------
#	Image URL
#-------------------------------------------------------------------------------

if (!preg_match("#^(https?://\S+)$#", $_GET['image'], $match)) {

	Fail ("Parameter 'image=' missing or unknown protocol");
}

$url = $match[1];

#-------------------------------------------------------------------------------
#	Checksum image url
#-------------------------------------------------------------------------------

$thumbHash = md5 ($url);

#-------------------------------------------------------------------------------
#	Cache entry
#-------------------------------------------------------------------------------

$cachePath = "{$CACHE_DIR}/{$thumbHash}.jpg";
$cacheURL  = "{$CACHE_DIR}/{$thumbHash}.url";

$message = "'$url': negative cached.";

if ($DEBUG >= 3) error_log("cachePath: $cachePath");

#-------------------------------------------------------------------------------
#	If not in cache, must fetch
#-------------------------------------------------------------------------------

$mustFetch = !file_exists ($cachePath) || $NO_CACHE;

if ($DEBUG >= 3) error_log("mustFetch: ". !!$mustFetch);

#-------------------------------------------------------------------------------
#	If in cache, check if stale
#-------------------------------------------------------------------------------

if (!$mustFetch) {

	#-------------------------------------------------------------------------------
	#	Determine max age for positive and negative cache entries
	#-------------------------------------------------------------------------------

	$maxAge = filesize($cachePath)>0 ? $IMAGE_POS_CACHE_TIME : $IMAGE_NEG_CACHE_TIME;

	#-------------------------------------------------------------------------------
	#	If too old, must fetch
	#-------------------------------------------------------------------------------

	if (time() - filemtime ($cachePath) > $maxAge) $mustFetch = TRUE;
}

#-------------------------------------------------------------------------------
#	Fetch if needed
#-------------------------------------------------------------------------------

if ($DEBUG >= 3) error_log("mustFetch: $mustFetch");

if ($mustFetch) {

	$message = getThumb ($url, $cachePath, $cacheURL);
	touch ($cachePath);
}

#-------------------------------------------------------------------------------
#	If still not in cache, throw error.
#-------------------------------------------------------------------------------

$size = @filesize ($cachePath);

if (!$size) Fail ($message);

#-------------------------------------------------------------------------------
#	Show messages instead of thumbnail if requested by DEBUG parameter.
#-------------------------------------------------------------------------------

if ($DEBUG>=2) {
	$stamp = date ('c', filemtime ($cachePath));
	Fail ("Thumbnail '{$thumbHash}' created at {$stamp} available.");
}

#-------------------------------------------------------------------------------
#	Components of HTTP header
#-------------------------------------------------------------------------------

$redirectURL = trim (@file_get_contents ($cacheURL));

$modtime = date ('r', filemtime ($cachePath));

#-------------------------------------------------------------------------------
#	Redirect if possible
#-------------------------------------------------------------------------------

if ($redirectURL) {

	header ("HTTP/1.0 302 Found");

	header ("Location: {$redirectURL}");
	header ("Last-Modified: {$modtime}");
	header ("Content-Type: text/html");

	$htmlRedirectURL = html_specialchars ($redirectURL);

	echo <<<EOT
<html>
<head>
<title>Redirect to {$htmlRedirectURL}</title>
</head>
<body>
<p>
Redirecting to <a href="${$htmlRedirectURL}">{$htmlRedirectURL}</a>.
</p>
</body>
</html>
EOT;
	exit ();
}

#-------------------------------------------------------------------------------
#	Deliver image from cache
#-------------------------------------------------------------------------------

header ("Content-Type: image/jpeg");
header ("Content-Length: {$size}");
header ("Last-Modified: {$modtime}");

readfile ($cachePath);

exit(0);

#-------------------------------------------------------------------------------
#	If webpage instead of image retrieved, try to extract image url.
#-------------------------------------------------------------------------------

function getThumb ($sourceURL, $cachePath, $cacheURL) {

	global $NO_CACHE;

	$thumbURL = '';

	for ($redirect=0; $redirect<3; $redirect++) {

		logMsg("URL:\t{$sourceURL}");

		#---------------------------------------------------------------
		#	Parse URL
		#---------------------------------------------------------------

		if (!preg_match ('!^(https?://)?(\w+(?:-\w+)*(?:\.\w+(?:-\w+)*)+(?::\d+)?)(/[^#?]*)?(\?[^#]*)?(#.*)?$!s',
			$sourceURL, $match)) {

			@unlink($cachePath);
			@unlink($cacheURL);

			return "Invalid URL";
		}

		$protocol= @$match[1] ? $match[1] : 'http://';
		$domain  = @$match[2];
		$path    = @$match[3] ? $match[3] : '/';
		$query   = @$match[4] ? $match[4] : '';
		$hash    = @$match[5] ? $match[5] : '';

		#---------------------------------------------------------------
		#	Access control
		#---------------------------------------------------------------

		usedTime();

		if (!checkAccess ($protocol, $domain, preg_replace ('#[^/]+$#', '', $path))) {

			@unlink($cachePath);
			@unlink($cacheURL);

			return "Access not allowed by robots.txt.";
		}

		logMsg(usedTime()."s\tused for access check.");

		#---------------------------------------------------------------
		#	Fetch content of URL
		#---------------------------------------------------------------

		$data = @file_get_contents ($sourceURL);

		if ($data === FALSE) {

			@unlink($cachePath);
			@unlink($cacheURL);

			logMsg(usedTime()."s\tused for failed retrieval.");

			return "Couldn't fetch URL '$sourceURL'.";
		}

		$size = strlen($data);

		logMsg(usedTime()."s\tused for retrieval of {$size} bytes.");

		#---------------------------------------------------------------
		#	Decode image.
		#---------------------------------------------------------------

		$image = @imagecreatefromstring ($data);

		logMsg(usedTime()."s\tused for decoding image.");

		#---------------------------------------------------------------
		#	If image then :-)
		#---------------------------------------------------------------

		if ($image !== FALSE) break;

		#---------------------------------------------------------------
		#	No image, no thumb.
		#---------------------------------------------------------------

		$thumbURL = '';

		#---------------------------------------------------------------
		#	OSM-Wiki and link to /w/images/ ?
		#---------------------------------------------------------------
		#src="/w/images/thumb/6/68/Stundenstein_Heideweg.jpg/800px-Stundenstein_Heideweg.jpg"
		#---------------------------------------------------------------

		if (preg_match ('#http://(wiki\.openstreetmap\.org)/#', $sourceURL, $host) &&
			preg_match ('#src="(/w/images/thumb/[^<>"]+)"#', $data, $match)) {

			$sourceURL = "http://{$host[1]}{$match[1]}";
			logMsg("Rule:\tLink to '/w/images/thumb/*' on wiki.openstreetmap.org page.");
			continue;
		}

		#---------------------------------------------------------------
		#	On commons page: img from //upload.wikimedia.org/* ?
		#---------------------------------------------------------------

		if (preg_match ('#//commons\.wikimedia\.org/wiki/File:#', $sourceURL) &&
			preg_match ('#src="(//upload.wikimedia.org/[^\\s"<>]*\bthumb/[^\\s"<>]*jpe?g)"#i', $data, $match)) {

			$sourceURL = "http:{$match[1]}";
			logMsg("Rule:\tImage from '//upload.wikimedia.org/*.jp[e]g' on commons page.");
			continue;
		}

		#---------------------------------------------------------------
		#	On any page: Link to //upload.wikimedia.org/* ?
		#---------------------------------------------------------------

		if (preg_match ('#href="(//upload.wikimedia.org/[^\\s"<>]+jpe?g)"#i', $data, $match)) {

			$sourceURL = "http:{$match[1]}";
			logMsg("Rule:\tLink to '//upload.wikimedia.org/*.jp[e]g' on any page.");
			continue;
		}

		#---------------------------------------------------------------
		#	Wikipedia page and link to /wiki/[Namespace}:* ?
		#---------------------------------------------------------------
		#	<a href="/wiki/{Namespace}:{filename}jp[e]g" class="image"
		#---------------------------------------------------------------

		if (preg_match ('#//(\w+\.wikipedia\.org)/#', $sourceURL, $host) &&
			preg_match ('#<a href="(/wiki/\w+:[^<>"]+jpe?g)"\s+class="image"[^>]*><img#i',
				$data, $match)) {

			$sourceURL = "http://{$host[1]}{$match[1]}";
			logMsg("Rule:\tLink to '/wiki/{Namespace}:{Name}.jp[e]g' on *.wikipedia.org page.");
			continue;
		}

		#---------------------------------------------------------------
		#	Wikipedia page and link to /w/index.php?title=[Namespace}:* ?
		#---------------------------------------------------------------
		#	<a href="/w/index.php?title={Namespace}:{filename}jp[e]g&..."
		#---------------------------------------------------------------

		if (preg_match ('#//(\w+\.wikipedia\.org)/#', $sourceURL, $host) &&
			preg_match ('#<a href="(/w/index.php\?title=[^<>"&]+jpe?g(?:&[^<>"]*)?)"#i',
				$data, $match)) {

			$sourceURL = "http://{$host[1]}" . htmlspecialchars_decode($match[1]);
			logMsg("Rule:\tLink to '/w/index.php?title={Namespace}:{Name}.jp[e]g' on *.wikipedia.org page.");
			continue;
		}

		#---------------------------------------------------------------
		#	Wikipedia page and srcset ?
		#---------------------------------------------------------------
		#	srcset="//upload.wikimedia.org/wikipedia/de/1/1a/Denkmal_sanitaetsrat_dr_baumann.jpg 
		#---------------------------------------------------------------

		if (preg_match ('#srcset="(//upload.wikimedia.org/[^\\s"<>]+jpe?g)\\s#i', $data, $match)) {

			$sourceURL = "http:{$match[1]}";
			logMsg("Rule:\tLink to '//upload.wikimedia.org/*.jp[e]g' as srcset on any page.");
			continue;
		}

		#---------------------------------------------------------------
		#	Flickr image page?
		#---------------------------------------------------------------
		#	http://www.flickr.com/photos/{USER}/{IDNUMBER}/?
		#---------------------------------------------------------------
		#	<noscript><img src="http://farm{DIGIT*}.staticflickr.com/{NUMBER}/{DECNUM}_{HEXNUM}_{CHAR}.jpg"
		#---------------------------------------------------------------

		if (preg_match ('#//www\.flickr\.com/photos/\w+/\d+(?:/.*)?$#', $sourceURL, $host) &&
			preg_match ('#<noscript><img src="((http://farm\d+\.staticflickr\.com/\d+/\d+_[a-f\d]+_)(\w)(\.jpg))[?"]#',
				$data, $match)) {

			global $FLICKR_THUMB_SIZE;

			$sourceURL= $match[1];
			$thumbURL = "{$match[2]}{$FLICKR_THUMB_SIZE}{$match[4]}";
			logMsg("Rule:\tFlickr image in flickr page.");
			continue;
		}

		#---------------------------------------------------------------
		#	If no image then :-(
		#---------------------------------------------------------------

		break;
	}

	#-----------------------------------------------------------------------
	#	No image ?
	#-----------------------------------------------------------------------

	if ($image === FALSE) {

		@unlink($cachePath);
		@unlink($cacheURL);

		return "No image found for URL '$sourceURL' ($size bytes).";
	}

	#-----------------------------------------------------------------------
	#	Scale image
	#-----------------------------------------------------------------------

	global $THUMB_WIDTH;
	global $THUMB_HEIGHT;
	global $THUMB_MINRATIO;
	global $THUMB_MAXRATIO;

	$origWidth = imagesx($image);
	$origHeight= imagesy($image);

	#-----------------------------------------------------------------------
	#	Trim if neccessary
	#-----------------------------------------------------------------------

	$widthTrim = max (0, round (($origWidth - $origHeight*$THUMB_MAXRATIO) / 2));
	$heightTrim= max (0, round (($origHeight- $origWidth /$THUMB_MINRATIO) / 2));

	$trimmedWidth = $origWidth - 2 * $widthTrim;
	$trimmedHeight= $origHeight- 2 * $heightTrim;

	#-----------------------------------------------------------------------
	#	Scale down into given box preserving w/h-ratio. NO scale up.
	#-----------------------------------------------------------------------

	$scale = min (1, $THUMB_WIDTH/$trimmedWidth, $THUMB_HEIGHT/$trimmedHeight);

	$thumbWidth = round ($scale * $trimmedWidth);
	$thumbHeight= round ($scale * $trimmedHeight);

	$thumb = imagecreatetruecolor($thumbWidth, $thumbHeight);

	imagecopyresampled ($thumb, $image,
		0, 0, $widthTrim, $heightTrim,
		$thumbWidth, $thumbHeight, $trimmedWidth, $trimmedHeight);

	logMsg(usedTime()."s\tused for scaling image.");

	#-----------------------------------------------------------------------
	#	Store thumbnail URL
	#-----------------------------------------------------------------------

	if ($thumbURL) {

		if (atomicWrite ($cacheURL, $thumbURL, "Ok:\tRedirect '{$thumbHash}' created.")) {
			logMsg("Ok:\tRedirect '{$thumbHash}' created.");
		}

	} else {
		@unlink ($cacheURL);
	}

	#-----------------------------------------------------------------------
	#	Store thumbnail
	#-----------------------------------------------------------------------

	$tempFile = "{$cachePath}." . getmypid() . ".tmp";
	if (imagejpeg ($thumb, $tempFile)) {
		@atomicRename ($tempFile, $cachePath);
	}
	@unlink ($tempFile);

	logMsg(usedTime()."s\tused for storing thumb.");

	return '';
}

#-------------------------------------------------------------------------------
#	Atomically write file
#-------------------------------------------------------------------------------

function atomicWrite ($file, $content, $message) {

	$tempFile = "{$file}." . getmypid() . ".tmp";

	$fh = @fopen ($tempFile, "wb");

	if ($fh && @fwrite ($fh, $content) && @fclose($fh)) {
		if (atomicRename ($tempFile, $file)) return TRUE;
	}
	@unlink ($tempFile);	
	return FALSE;
}

function atomicRename ($tempFile, $destFile) {

	@chmod 	($tempFile, 0444);
	frename ($tempFile, $destFile);
	@unlink ($tempFile);
	return TRUE;
}

#-------------------------------------------------------------------------------
#	Check access by robot.txt
#-------------------------------------------------------------------------------

function checkAccess ($protocol, $domain, $directory) {

	global $ALLOWED_DOMAINS;
	global $CACHE_DIR;
	global $ROBOT_FRESH;

	foreach ($ALLOWED_DOMAINS as $allowedDomain) {

		$length = strlen ($allowedDomain);

		if (       $domain               ==    $allowedDomain) return TRUE;
		if (substr($domain, -($length+1))=='.'.$allowedDomain) return TRUE;
	}

	for ($path=$directory; $path; $path = preg_replace('#[^/]*/*[^/]*$#', '', $path)) {

		$url	= "{$protocol}{$domain}{$path}robots.txt";
		$hash	= md5($url);
		$cache	= "{$CACHE_DIR}/{$hash}.acc";
		$mtime	= @filemtime($cache);

		#---------------------------------------------------------------
		#	If outdated, read robot file
		#---------------------------------------------------------------

		if ($mtime < $ROBOT_FRESH) {

			#-------------------------------------------------------
			#	Fetch content of robots.txt
			#-------------------------------------------------------

			$lines = @file($url);

			$temp = $cache . '.' . getmypid() . '.tmp';
			$fh = @fopen ($temp, "w");

			if (!$fh) return FALSE;

			fwrite ($fh, "url\t{$url}\n");

			if ($lines!==FALSE) {

				$itsme	= FALSE;
				$result	= array();

				foreach ($lines as $line) {

					#---------------------------------------
					#	Check if applyable
					#---------------------------------------

					if (preg_match ('#^User-Agent:\s*(\S+)\s*$#i', $line, $match)) {

						$itsme = !strcasecmp ($match[1], 'OpenStreetMap');
						continue;
					}

					if (!$itsme) continue;

					#---------------------------------------
					#	Process
					#---------------------------------------

					if (preg_match ('#^(Allow|DisAllow):\s*(/\S*)\s*$#i', $line, $match)) {

						$result[$match[2]] = strtolower($match[1])=='allow' ?
							'allow' : 'deny';
						continue;
					}
				}

				#-------------------------------------------------------
				#	Order by string length
				#-------------------------------------------------------

				uksort ($result, orderbylength);

				foreach ($result as $prefix => $command) {

					fwrite ($fh, "{$command}\t{$prefix}\n");
				}
			}

			fclose ($fh);

			frename ($temp, $cache);
		}

		#---------------------------------------------------------------
		#	Check if allowed
		#---------------------------------------------------------------

		$rules = @file($cache);

		if ($rules === FALSE) continue;

		foreach ($rules as $rule) {

			$fields = explode ("\t", trim ($rule));

			if ($fields[0]=='allow' && strpos($directory, $fields[1])===0) return TRUE;
			if ($fields[0]=='deny'  && strpos($directory, $fields[1])===0) return FALSE;
		}
	}

	return FALSE;
}

#-------------------------------------------------------------------------------
#	Helper
#-------------------------------------------------------------------------------

function frename ($temp, $file) {

	if (!@rename ($temp, $file)) {
		@unlink ($file);
		@rename ($temp, $file);
	}
	@unlink ($temp);
}

function orderbylength ($a, $b) {

	$diff = strlen($b) - strlen($a);
	return $diff ? $diff : strcmp ($a, $b);
}

#===============================================================================
#	$Id: thumbcache.php,v 1.18 2013/04/21 17:59:18 wolf Exp $
#===============================================================================
?>
